from enum import Enum
from pydantic import BaseModel, Field
from .entity import Permission, Group, Member, Friend
from .message import MessageChain
from typing import Optional
from datetime import datetime


class EventTypes(Enum):
    """
    EventTypes is used for registering handlers
    """
    # Bot events
    BotOnlineEvent = 'BotOnlineEvent'
    BotOfflineEventActive = 'BotOfflineEventActive'
    BotOfflineEventForce = 'BotOfflineEventForce'
    BotOfflineEventDropped = 'BotOfflineEventDropped'
    BotReloginEvent = 'BotReloginEvent'
    BotGroupPermissionChangeEvent = 'BotGroupPermissionChangeEvent'
    BotMuteEvent = 'BotMuteEvent'
    BotUnmuteEvent = 'BotUnmuteEvent'
    BotJoinGroupEvent = 'BotJoinGroupEvent'

    # Group events
    GroupNameChangeEvent = 'GroupNameChangeEvent'
    GroupEntranceAnnouncementChangeEvent = 'GroupEntranceAnnouncementChangeEvent'
    GroupMuteAllEvent = 'GroupMuteAllEvent'
    GroupRecallEvent = "GroupRecallEvent"
    FriendRecallEvent = "FriendRecallEvent"

    # Group permission events
    GroupAllowAnonymousChatEvent = 'GroupAllowAnonymousChatEvent'  # 群设置 是否允许匿名聊天 被修改
    GroupAllowConfessTalkEvent = 'GroupAllowConfessTalkEvent'  # 坦白说
    GroupAllowMemberInviteEvent = 'GroupAllowMemberInviteEvent'  # 邀请进群

    # Group member events
    MemberJoinEvent = 'MemberJoinEvent'
    MemberLeaveEventKick = 'MemberLeaveEventKick'
    MemberLeaveEventQuit = 'MemberLeaveEventQuit'
    MemberCardChangeEvent = 'MemberCardChangeEvent'
    MemberSpecialTitleChangeEvent = 'MemberSpecialTitleChangeEvent'
    MemberPermissionChangeEvent = 'MemberPermissionChangeEvent'
    MemberMuteEvent = 'MemberMuteEvent'
    MemberUnmuteEvent = 'MemberUnmuteEvent'

    # Requests
    NewFriendRequestEvent = "NewFriendRequestEvent"
    MemberJoinRequestEvent = "MemberJoinRequestEvent"

    # Message events
    FriendMessage = 'FriendMessage'
    GroupMessage = 'GroupMessage'


class NewFriendRequestResponse(Enum):
    ACCEPT = 0
    REFUSE = 1
    REFUSE_AND_BLACKLIST = 2


class MemberJoinRequestResponse(Enum):
    ACCEPT = 0
    REFUSE = 1
    IGNORE = 2
    REFUSE_AND_BLACKLIST = 3
    IGNORE_AND_BLACKLIST = 4


class Event(BaseModel):
    type: EventTypes


class BotOnlineEvent(Event):
    type = EventTypes.BotOnlineEvent
    qq: int

    def __str__(self):
        return f'[{self.type.value}: qq={self.qq}]'


class BotOfflineEventActive(Event):
    type = EventTypes.BotOfflineEventActive
    qq: int

    def __str__(self):
        return f'[{self.type.value}: qq={self.qq}]'


class BotOfflineEventForce(Event):
    type = EventTypes.BotOfflineEventForce
    qq: int

    def __str__(self):
        return f'[{self.type.value}: qq={self.qq}]'


class BotOfflineEventDropped(Event):
    type = EventTypes.BotOfflineEventDropped
    qq: int

    def __str__(self):
        return f'[{self.type.value}: qq={self.qq}]'


class BotReloginEvent(Event):
    type = EventTypes.BotReloginEvent
    qq: int

    def __str__(self):
        return f'[{self.type.value}: qq={self.qq}]'


class BotGroupPermissionChangeEvent(Event):
    type = EventTypes.BotGroupPermissionChangeEvent
    origin: Permission
    current: Permission
    group: Group

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'origin={repr(self.origin)}, new={repr(self.current)}, group={repr(self.group)}]'


class BotMuteEvent(Event):
    type = EventTypes.BotMuteEvent
    durationSeconds: int
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type.value}: durationSeconds={self.durationSeconds}, operator={repr(self.operator)}]'


class BotUnmuteEvent(Event):
    type = EventTypes.BotUnmuteEvent
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type.value}: operator={repr(self.operator)}]'


class BotJoinGroupEvent(Event):
    type = EventTypes.BotJoinGroupEvent
    group: Group

    def __str__(self):
        return f'[{self.type.value}: group={repr(self.group)}]'


class GroupRecallEvent(Event):
    type: EventTypes = EventTypes.GroupRecallEvent
    authorId: int
    messageId: int
    time: datetime
    group: Group
    operator: Optional[Member]


class FriendRecallEvent(Event):
    type: EventTypes = EventTypes.FriendRecallEvent
    authorId: int
    messageId: int
    time: int
    operator: int


class GroupNameChangeEvent(Event):
    type = EventTypes.GroupNameChangeEvent
    origin: str
    current: str
    group: Group
    isByBot: bool

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'origin={self.origin}, new={self.current}, group={repr(self.group)}, isByBot={self.isByBot}]'


class GroupEntranceAnnouncementChangeEvent(Event):
    type = EventTypes.GroupEntranceAnnouncementChangeEvent
    origin: str
    current: str
    group: Group
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'origin={self.origin}, new={self.current}, group={repr(self.group)}, operator={repr(self.operator)}]'


class GroupMuteAllEvent(Event):
    type = EventTypes.GroupMuteAllEvent
    origin: bool
    current: bool
    group: Group
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'origin={self.origin}, new={self.current}, group={repr(self.group)}, operator={repr(self.operator)}]'


class GroupAllowAnonymousChatEvent(Event):
    type = EventTypes.GroupAllowAnonymousChatEvent
    origin: bool
    current: bool
    group: Group
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'origin={self.origin}, new={self.current}, group={repr(self.group)}, operator={repr(self.operator)}]'


class GroupAllowConfessTalkEvent(Event):
    type = EventTypes.GroupAllowAnonymousChatEvent
    origin: bool
    current: bool
    group: Group
    isByBot: bool

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'origin={self.origin}, new={self.current}, group={repr(self.group)}, isByBot={self.isByBot}]'


class GroupAllowMemberInviteEvent(Event):
    type = EventTypes.GroupAllowMemberInviteEvent
    origin: bool
    current: bool
    group: Group
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'origin={self.origin}, new={self.current}, group={repr(self.group)}, operator={repr(self.operator)}]'


class MemberJoinEvent(Event):
    type = EventTypes.MemberJoinEvent
    member: Member

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'member={repr(self.member)}]'


class MemberLeaveEventKick(Event):
    type = EventTypes.MemberLeaveEventKick
    member: Member
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'member={repr(self.member)}, operator={repr(self.operator)}]'


class MemberLeaveEventQuit(Event):
    type = EventTypes.MemberLeaveEventQuit
    member: Member

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'member={repr(self.member)}]'


class MemberCardChangeEvent(Event):
    type = EventTypes.MemberCardChangeEvent
    origin: str
    current: str
    member: Member
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'origin={self.origin}, new={self.current}, member={repr(self.member)}, operator={repr(self.operator)}]'


class MemberSpecialTitleChangeEvent(Event):
    type = EventTypes.MemberSpecialTitleChangeEvent
    origin: str
    current: str
    member: Member

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'origin={self.origin}, new={self.current}, member={repr(self.member)}]'


class MemberPermissionChangeEvent(Event):
    type = EventTypes.MemberPermissionChangeEvent
    origin: str
    current: str
    member: Member

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'origin={self.origin}, new={self.current}, member={repr(self.member)}]'


class MemberMuteEvent(Event):
    type = EventTypes.MemberMuteEvent
    durationSeconds: int
    member: Member
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'durationSeconds={self.durationSeconds}, member={repr(self.member)}, operator={repr(self.operator)}]'


class MemberUnmuteEvent(Event):
    type = EventTypes.MemberUnmuteEvent
    member: Member
    operator: Optional[Member]

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'member={repr(self.member)}, operator={repr(self.operator)}]'


class FriendMessage(Event):
    type: EventTypes = EventTypes.FriendMessage
    messageChain: Optional[MessageChain]
    sender: Friend

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'messageChain={str(self.messageChain)}, sender={repr(self.sender)}]'


class GroupMessage(Event):
    type: EventTypes = EventTypes.GroupMessage
    messageChain: Optional[MessageChain]
    sender: Member

    def __str__(self):
        return f'[{self.type.value}: ' \
               f'messageChain={str(self.messageChain)}, sender={repr(self.sender)}]'


class NewFriendRequestEvent(Event):
    type: EventTypes = EventTypes.NewFriendRequestEvent
    requestId: int = Field(..., alias="eventId")
    supplicant: int = Field(..., alias="fromId")  # 即请求方 QQ
    sourceGroup: Optional[int] = Field(..., alias="groupId")
    nickname: str = Field(..., alias="nick")


class MemberJoinRequestEvent(Event):
    type: EventTypes = EventTypes.MemberJoinRequestEvent
    requestId: int = Field(..., alias="eventId")
    supplicant: int = Field(..., alias="fromId")  # 即请求方 QQ
    sourceGroup: Optional[int] = Field(..., alias="groupId")
    groupName: str = Field(..., alias="groupName")
    nickname: str = Field(..., alias="nick")


class Events(Enum):
    """
    Internal use only
    """
    BotOnlineEvent = BotOnlineEvent
    BotOfflineEventActive = BotOfflineEventActive
    BotOfflineEventForce = BotOfflineEventForce
    BotOfflineEventDropped = BotOfflineEventDropped
    BotReloginEvent = BotReloginEvent
    BotGroupPermissionChangeEvent = BotGroupPermissionChangeEvent
    BotMuteEvent = BotMuteEvent
    BotUnmuteEvent = BotUnmuteEvent
    BotJoinGroupEvent = BotJoinGroupEvent

    GroupNameChangeEvent = GroupNameChangeEvent
    GroupEntranceAnnouncementChangeEvent = GroupEntranceAnnouncementChangeEvent
    GroupMuteAllEvent = GroupMuteAllEvent
    GroupRecallEvent = GroupRecallEvent
    FriendRecallEvent = FriendRecallEvent

    # 群设置被修改事件
    GroupAllowAnonymousChatEvent = GroupAllowAnonymousChatEvent  # 群设置 是否允许匿名聊天 被修改
    GroupAllowConfessTalkEvent = GroupAllowConfessTalkEvent  # 坦白说
    GroupAllowMemberInviteEvent = GroupAllowMemberInviteEvent  # 邀请进群

    # 群事件(被 Bot 监听到的, 为被动事件, 其中 Bot 身份为第三方.)
    MemberJoinEvent = MemberJoinEvent
    MemberLeaveEventKick = MemberLeaveEventKick
    MemberLeaveEventQuit = MemberLeaveEventQuit
    MemberCardChangeEvent = MemberCardChangeEvent
    MemberSpecialTitleChangeEvent = MemberSpecialTitleChangeEvent
    MemberPermissionChangeEvent = MemberPermissionChangeEvent
    MemberMuteEvent = MemberMuteEvent
    MemberUnmuteEvent = MemberUnmuteEvent

    # Join
    NewFriendRequestEvent = NewFriendRequestEvent
    MemberJoinRequestEvent = MemberJoinRequestEvent

    FriendMessage = FriendMessage
    GroupMessage = GroupMessage



from pydantic import BaseModel
from typing import Optional
from enum import Enum


class Friend(BaseModel):
    id: int
    nickname: Optional[str]
    remark: Optional[str]

    def __repr__(self):
        return f"<Friend id={self.id} nickname='{self.nickname}' remark='{self.remark}'>"

    def get_avatar_url(self) -> str:
        return f'http://q4.qlogo.cn/g?b=qq&nk={self.id}&s=140'


class Permission(Enum):
    Member = "MEMBER"
    Administrator = "ADMINISTRATOR"
    Owner = "OWNER"


class Group(BaseModel):
    id: int
    name: str
    permission: Permission

    def __repr__(self):
        return f"<Group id={self.id} name='{self.name}' permission={self.permission.name}>"

    def get_avatar_url(self) -> str:
        return f'https://p.qlogo.cn/gh/{self.id}/{self.id}/'


class Member(BaseModel):
    id: int
    memberName: str
    permission: Permission
    group: Group

    def __repr__(self):
        return f"<GroupMember id={self.id} group={self.group} permission={self.permission} group={self.group.id}>"

    def get_avatar_url(self) -> str:
        return f'http://q4.qlogo.cn/g?b=qq&nk={self.id}&s=140'


class MemberChangeableSetting(BaseModel):
    name: str
    specialTitle: str

    def modify(self, **kwargs):
        for i in ("name", "kwargs"):
            if i in kwargs:
                setattr(self, i, kwargs[i])
        return self


class GroupSetting(BaseModel):
    name: str
    announcement: str
    confessTalk: bool
    allowMemberInvite: bool
    autoApprove: bool
    anonymousChat: bool

    def modify(self, **kwargs):
        for i in ("name",
                  "announcement",
                  "confessTalk",
                  "allowMemberInvite",
                  "autoApprove",
                  "anonymousChat"
                  ):
            if i in kwargs:
                setattr(self, i, kwargs[i])
        return self



from typing import Dict
import aiohttp
from aiohttp import client_exceptions
from .log import create_logger
from io import BytesIO

from .exceptions import AuthenticationException, NetworkException, ServerException, \
    UnknownTargetException, PrivilegeException, BadRequestException, MiraiException, SessionException


error_code = {
                1: AuthenticationException('Incorrect authKey'),
                2: AuthenticationException('Bot does not exist'),
                3: SessionException('Session does not exist or has expired'),
                4: AuthenticationException('Session is not verified'),
                5: UnknownTargetException('Message target does not exist'),
                10: PrivilegeException('Bot does not have corresponding privilege'),
                400: BadRequestException('Bad Request, please check arguments/url'),
            }


class HttpClient:
    """
    Internal use only
    HttpClient implemented by aiohttp
    """

    DEFAULT_TIMEOUT = 5

    @staticmethod
    async def _check_response(result: aiohttp.ClientResponse, url, method) -> Dict:
        """
        Check url response, and raise exceptions

        :param result: http response
        :param url: url to show in the log
        :param method: 'post', 'get'
        :return: json decoded result
        """
        if result.status != 200:
            raise ServerException(f'{url} {method} failed, status code: {result.status}')
        result = await result.json()
        if not isinstance(result, dict):
            return result
        status_code = result.get('code')
        if method == 'post':
            if status_code is None:
                raise ServerException('Empty response')
            if status_code == 0:  # normal
                return result
        elif method == 'get':
            if status_code is None or status_code == 0:
                return result
        if status_code in error_code:
            raise error_code[status_code]
        else:
            raise MiraiException('HTTP API updated, please upgrade python-mirai-core')

    def __init__(self, base_url: str, timeout=DEFAULT_TIMEOUT, loop=None):
        self.base_url = base_url
        self.timeout = aiohttp.ClientTimeout(timeout)
        self.session = aiohttp.ClientSession(timeout=self.timeout, loop=loop)
        self.logger = create_logger('Network')
        self.loop = loop

    async def get(self, url, headers=None, params=None):
        """
        send http get request

        :param url: the sub url
        :param headers: request headers
        :param params: get params
        :return: json decoded response
        """
        if url != '/fetchMessage':
            self.logger.debug(f'get {url} with params: {str(params)}')
        try:
            response = await self.session.get(self.base_url + url, headers=headers, params=params)
        except client_exceptions.ClientConnectorError:
            raise NetworkException('Unable to reach Mirai console')
        return await HttpClient._check_response(response, url, 'get')

    async def post(self, url, headers=None, data=None):
        """
        send http post request

        :param url: the sub url
        :param headers: request headers
        :param data: post params
        :return: json decoded response
        """

        self.logger.debug(f'post {url} with data: {str(data)}')
        try:
            response = await self.session.post(self.base_url + url, headers=headers, json=data)
        except client_exceptions.ClientConnectorError:
            raise NetworkException('Unable to reach Mirai console')
        return await HttpClient._check_response(response, url, 'post')

    async def upload(self, url, headers=None, data=None, file: str = None):
        """
        upload using multipart upload

        :param url: the sub url
        :param headers: request headers
        :param data: post params
        :param file: file to attach
        :return: json decoded response
        """
        if data is None:
            data = dict()
        data['img'] = BytesIO(open(str(file.absolute()), 'rb').read())
        self.logger.debug(f'upload {url} with file: {file}')
        try:
            response = await self.session.post(self.base_url + url,
                                               headers=headers, data=data)
        except client_exceptions.ClientConnectorError:
            raise NetworkException('Unable to reach Mirai console')
        self.logger.debug(f'Image uploaded: {response.text}')
        return await response.json()

    async def websocket(self, url: str, handler: callable, ws_close_handler: callable):
        """
        Create websocket subscriber to url

        :param url: the sub url
        :param handler: request headers
        :param ws_close_handler: callback for connection close
        """
        try:
            ws = await self.session.ws_connect(self.base_url + url)
            self.logger.debug('Websocket established')
            while True:
                msg = await ws.receive()
                if msg.type == aiohttp.WSMsgType.TEXT:
                    self.logger.debug(f'Websocket received {msg}')
                    await handler(msg.json())
                elif msg.type == aiohttp.WSMsgType.CLOSED:
                    self.logger.debug('Websocket closed')
                    await ws_close_handler()
                    return
                else:
                    self.logger.warning(f'Received unexpected type: {msg.type}')
        except client_exceptions.ClientConnectorError:
            raise NetworkException('Unable to reach Mirai console')

    async def close(self):
        """
        Close session
        """
        await self.session.close()



from enum import Enum
from pydantic import BaseModel
from abc import abstractmethod
from enum import Enum
from typing import List, Dict, Optional, overload, Iterable, Union
from pydantic import Field, validator, HttpUrl, BaseModel
from pydantic.generics import GenericModel
from pathlib import Path
import datetime
from collections import MutableSequence

__all__ = [
    "MessageComponentTypes",
    "BaseMessageComponent",
    'Plain',
    'Source',
    'At',
    'AtAll',
    'Face',
    'Image',
    'Unknown',
    'Quote',
    'ComponentTypes',
    'LocalImage',
    'MessageChain',
    'qq_emoji_text_list'
]


class MessageComponentTypes(Enum):
    Source = "Source"
    Plain = "Plain"
    Face = "Face"
    At = "At"
    AtAll = "AtAll"
    Image = "Image"
    Quote = "Quote"
    Xml = "Xml"
    Json = "Json"
    App = "App"
    Unknown = "Unknown"


class BaseMessageComponent(BaseModel):
    type: MessageComponentTypes

    def __str__(self):
        """
        for plain text extraction
        :return: human readable text component
        """
        return ''


# original text copied from Tim
qq_emoji_text_list = {
    0:   '[惊讶]',
    1:   '[撇嘴]',
    2:   '[色]',
    3:   '[发呆]',
    4:   '[得意]',
    5:   '[流泪]',
    6:   '[害羞]',
    7:   '[闭嘴]',
    8:   '[睡]',
    9:   '[大哭]',
    10:  '[尴尬]',
    11:  '[发怒]',
    12:  '[调皮]',
    13:  '[呲牙]',
    14:  '[微笑]',
    15:  '[难过]',
    16:  '[酷]',
    17:  '[Empty]',
    18:  '[抓狂]',
    19:  '[吐]',
    20:  '[偷笑]',
    21:  '[可爱]',
    22:  '[白眼]',
    23:  '[傲慢]',
    24:  '[饥饿]',
    25:  '[困]',
    26:  '[惊恐]',
    27:  '[流汗]',
    28:  '[憨笑]',
    29:  '[悠闲]',
    30:  '[奋斗]',
    31:  '[咒骂]',
    32:  '[疑问]',
    33:  '[嘘]',
    34:  '[晕]',
    35:  '[折磨]',
    36:  '[衰]',
    37:  '[骷髅]',
    38:  '[敲打]',
    39:  '[再见]',
    40:  '[Empty]',
    41:  '[发抖]',
    42:  '[爱情]',
    43:  '[跳跳]',
    44:  '[Empty]',
    45:  '[Empty]',
    46:  '[猪头]',
    47:  '[Empty]',
    48:  '[Empty]',
    49:  '[拥抱]',
    50:  '[Empty]',
    51:  '[Empty]',
    52:  '[Empty]',
    53:  '[蛋糕]',
    54:  '[闪电]',
    55:  '[炸弹]',
    56:  '[刀]',
    57:  '[足球]',
    58:  '[Empty]',
    59:  '[便便]',
    60:  '[咖啡]',
    61:  '[饭]',
    62:  '[Empty]',
    63:  '[玫瑰]',
    64:  '[凋谢]',
    65:  '[Empty]',
    66:  '[爱心]',
    67:  '[心碎]',
    68:  '[Empty]',
    69:  '[礼物]',
    70:  '[Empty]',
    71:  '[Empty]',
    72:  '[Empty]',
    73:  '[Empty]',
    74:  '[太阳]',
    75:  '[月亮]',
    76:  '[赞]',
    77:  '[踩]',
    78:  '[握手]',
    79:  '[胜利]',
    80:  '[Empty]',
    81:  '[Empty]',
    82:  '[Empty]',
    83:  '[Empty]',
    84:  '[Empty]',
    85:  '[飞吻]',
    86:  '[怄火]',
    87:  '[Empty]',
    88:  '[Empty]',
    89:  '[西瓜]',
    90:  '[Empty]',
    91:  '[Empty]',
    92:  '[Empty]',
    93:  '[Empty]',
    94:  '[Empty]',
    95:  '[Empty]',
    96:  '[冷汗]',
    97:  '[擦汗]',
    98:  '[抠鼻]',
    99:  '[鼓掌]',
    100: '[糗大了]',
    101: '[坏笑]',
    102: '[左哼哼]',
    103: '[右哼哼]',
    104: '[哈欠]',
    105: '[鄙视]',
    106: '[委屈]',
    107: '[快哭了]',
    108: '[阴险]',
    109: '[亲亲]',
    110: '[吓]',
    111: '[可怜]',
    112: '[菜刀]',
    113: '[啤酒]',
    114: '[篮球]',
    115: '[乒乓]',
    116: '[示爱]',
    117: '[瓢虫]',
    118: '[抱拳]',
    119: '[勾引]',
    120: '[拳头]',
    121: '[差劲]',
    122: '[爱你]',
    123: '[NO]',
    124: '[OK]',
    125: '[转圈]',
    126: '[磕头]',
    127: '[回头]',
    128: '[跳绳]',
    129: '[挥手]',
    130: '[激动]',
    131: '[街舞]',
    132: '[献吻]',
    133: '[左太极]',
    134: '[右太极]',
    135: '[Empty]',
    136: '[双喜]',
    137: '[鞭炮]',
    138: '[灯笼]',
    139: '[发财]',
    140: '[K歌]',
    141: '[购物]',
    142: '[邮件]',
    143: '[帅]',
    144: '[喝彩]',
    145: '[祈祷]',
    146: '[爆筋]',
    147: '[棒棒糖]',
    148: '[喝奶]',
    149: '[下面]',
    150: '[香蕉]',
    151: '[飞机]',
    152: '[开车]',
    153: '[高铁左车头]',
    154: '[车厢]',
    155: '[高铁右车头]',
    156: '[多云]',
    157: '[下雨]',
    158: '[钞票]',
    159: '[熊猫]',
    160: '[灯泡]',
    161: '[风车]',
    162: '[闹钟]',
    163: '[打伞]',
    164: '[彩球]',
    165: '[钻戒]',
    166: '[沙发]',
    167: '[纸巾]',
    168: '[药]',
    169: '[手枪]',
    170: '[青蛙]',
    171: '[茶]',
    172: '[眨眼睛]',
    173: '[泪奔]',
    174: '[无奈]',
    175: '[卖萌]',
    176: '[小纠结]',
    177: '[喷血]',
    178: '[斜眼笑]',
    179: '[doge]',
    180: '[惊喜]',
    181: '[骚扰]',
    182: '[笑哭]',
    183: '[我最美]',
    184: '[河蟹]',
    185: '[羊驼]',
    186: '[Empty]',
    187: '[幽灵]',
    188: '[蛋]',
    189: '[Empty]',
    190: '[菊花]',
    191: '[Empty]',
    192: '[红包]',
    193: '[大笑]',
    194: '[不开心]',
    195: '[Empty]',
    196: '[Empty]',
    197: '[冷漠]',
    198: '[呃]',
    199: '[好棒]',
    200: '[拜托]',
    201: '[点赞]',
    202: '[无聊]',
    203: '[托脸]',
    204: '[吃]',
    205: '[送花]',
    206: '[害怕]',
    207: '[花痴]',
    208: '[小样儿]',
    209: '[Empty]',
    210: '[飙泪]',
    211: '[我不看]',
    212: '[托腮]',
    213: '[Empty]',
    214: '[啵啵]',
    215: '[糊脸]',
    216: '[拍头]',
    217: '[扯一扯]',
    218: '[舔一舔]',
    219: '[蹭一蹭]',
    220: '[拽炸天]',
    221: '[顶呱呱]',
    222: '[抱抱]',
    223: '[暴击]',
    224: '[开枪]',
    225: '[撩一撩]',
    226: '[拍桌]',
    227: '[拍手]',
    228: '[恭喜]',
    229: '[干杯]',
    230: '[嘲讽]',
    231: '[哼]',
    232: '[佛系]',
    233: '[掐一掐]',
    234: '[惊呆]',
    235: '[颤抖]',
    236: '[啃头]',
    237: '[偷看]',
    238: '[扇脸]',
    239: '[原谅]',
    240: '[喷脸]',
    241: '[生日快乐]',
    242: '[Empty]',
    243: '[Empty]',
    244: '[Empty]',
    245: '[Empty]',
    246: '[Empty]',
    247: '[Empty]',
    248: '[Empty]',
    249: '[Empty]',
    250: '[Empty]',
    251: '[Empty]',
    252: '[Empty]',
    253: '[Empty]',
    254: '[Empty]',
    255: '[Empty]',
}


class MessageChain(BaseModel, MutableSequence):
    # stores the actual components
    __root__: List[BaseMessageComponent] = []

    def insert(self, index: int, object) -> None:
        self.__root__.insert(index, object)

    @overload
    @abstractmethod
    def __setitem__(self, i: int, o) -> None:
        ...

    @overload
    @abstractmethod
    def __setitem__(self, s: slice, o: Iterable) -> None:
        ...

    def __setitem__(self, i: int, o) -> None:
        self.__root__.__setitem__(i, o)

    @overload
    @abstractmethod
    def __delitem__(self, i: int) -> None:
        ...

    @overload
    @abstractmethod
    def __delitem__(self, i: slice) -> None:
        ...

    def __delitem__(self, i: int) -> None:
        self.__root__.__delitem__(i)

    def __add__(self, value):
        # merge two message chain or append one component
        if isinstance(value, BaseMessageComponent):
            self.__root__.append(value)
            return self
        elif isinstance(value, MessageChain):
            self.__root__ += value.__root__
            return self

    def __str__(self) -> str:
        return ''.join([str(i) for i in self.__root__])

    @classmethod
    def custom_parse(cls, value: List[Dict]) -> 'MessageChain':
        """
        construct message chain from dict
        used only when receiving messages from mirai

        :param value: dict contains message components
        :return: MessageChain
        """
        for i in value:
            if not isinstance(i, dict):
                raise TypeError("invaild value")
        return cls(__root__=[ComponentTypes.__members__[m['type']].value(**m) for m in value])

    def __iter__(self):
        return self.__root__.__iter__()

    def __getitem__(self, index):
        return self.__root__.__getitem__(index)

    def has(self, component_class) -> bool:
        """
        test if any item in MessageChain is component_class

        :param component_class: the class for the component
        :return: boolean
        """
        for i in self:
            if isinstance(i, component_class):
                return True
        else:
            return False

    def __len__(self) -> int:
        return len(self.__root__)

    def get_first(self, component_class) -> Optional[BaseMessageComponent]:
        """
        Get the first component with component_class

        :param component_class: the class for the component
        :return: None or the component
        """
        for i in self.__root__:
            if isinstance(i, component_class):
                return i

    def get_all(self, component_class) -> List[BaseMessageComponent]:
        return [i for i in self.__root__ if isinstance(i, component_class)]

    def get_source(self) -> 'Source':
        result = self.get_first(Source)
        assert isinstance(result, Source)
        return result

    def get_quote(self) -> 'Quote':
        result = self.get_first(Quote)
        if result:
            assert isinstance(result, Quote)
            return result


class Source(BaseMessageComponent):
    """
    The source of the message
    Not a valid component for outbound message
    """
    type = MessageComponentTypes.Source
    id: int  # the message id: higher 32 bits are sequence id, and lower 32 bits are random id
    time: datetime.datetime  # the timestamp the message was sent

    def __str__(self):
        return ''

    def __repr__(self):
        return f'[Source: id={self.id}, time={self.time}]'


class Plain(BaseMessageComponent):
    """
    Text component
    Available for outbound message
    """
    type = MessageComponentTypes.Plain
    text: str  # text content

    def __init__(self, text: str, **kwargs):
        """
        Construct text component

        :param text: message text
        """
        super().__init__(text=text, **kwargs)

    def __str__(self):
        return self.text

    def __repr__(self):
        return f'[Plain: {self.text}]'


class Quote(BaseMessageComponent):
    """
    Quote component
    Not a valid component for outbound message
    """
    type = MessageComponentTypes.Quote
    id: int  # origin message id
    groupId: int  # origin group id
    senderId: int  # origin sender id
    origin: MessageChain  # origin message (Source and other content without Quote)

    def __str__(self):
        return ''

    def __repr__(self):
        return f'[Quote: id={self.id}]'


class At(GenericModel, BaseMessageComponent):
    """
    At component
    Available for outbound message
    """
    type = MessageComponentTypes.At
    target: int  # target qq number
    display: str  # display name (only used for inbound at), '@' already included

    def __init__(self, target, **kwargs):
        """
        Construct at component

        :param target: target qq number
        """
        super().__init__(target=target, **kwargs)

    def __str__(self):
        return self.display

    def __repr__(self):
        return f'[At: target={self.target}, display={self.display}]'


class AtAll(BaseMessageComponent):
    """
    AtAll component
    Available for outbound message
    """
    type = MessageComponentTypes.AtAll

    def __str__(self):
        return '@All'

    def __repr__(self):
        return f'[AtAll]'


class Face(BaseMessageComponent):
    """
    Face component
    Available for outbound message
    """
    type = MessageComponentTypes.Face
    faceId: int  # face id, see qq_emoji_text_list for details

    def __init__(self, faceId, **kwargs):
        """
        Construct Face component

        :param face_id: unsigned 8-bit face id
        """
        super().__init__(faceId=faceId, **kwargs)

    def __str__(self):
        return qq_emoji_text_list[self.faceId]

    def __repr__(self):
        return f'[Face: id={self.faceId}, {qq_emoji_text_list[self.faceId]}]'


class Image(BaseMessageComponent):
    """
    Received image component
    Available for outbound message
    """
    type = MessageComponentTypes.Image
    imageId: str
    url: Optional[HttpUrl] = None

    def __init__(self, imageId, **kwargs):
        """
        Construct Image from mirai styled uuid

        :param imageId: uuid as str (see https://github.com/mamoe/mirai-api-http/blob/master/MessageType.md#image)
        """
        super().__init__(imageId=imageId, **kwargs)

    def __str__(self):
        return ''

    def __repr__(self):
        return f'[Image: {self.imageId}]'

    @property
    def image_type(self):
        """
        get image type from pattern

        :return: 'friend' or 'group', if uuid is invalid format, return 'unknown'
        """
        if self.imageId.startswith('/'):
            return 'friend'
        elif self.imageId.startswith('{'):
            return 'group'
        else:
            'unknown'


class LocalImage:
    """
    Local image component
    Available for outbound message
    """
    path: Path  # absolute path to image

    def __init__(self, path=None):
        """
        Construct LocalImage component

        :param path: absolute path on disk
        """
        if isinstance(path, str):
            self.path = Path(path)
        elif isinstance(path, Path):
            self.path = path


class Xml(BaseMessageComponent):
    """
    Xml component
    Available for outbound message
    """
    type = MessageComponentTypes.Xml
    XML: str  # xml content

    def __init__(self, XML: str, **kwargs):
        """
        Construct Xml component

        :param xml: str contains xml
        """
        super().__init__(XML=XML, **kwargs)


class Json(BaseMessageComponent):
    """
    Json component
    Available for outbound message
    """
    type = MessageComponentTypes.Json
    Json: dict = Field(..., alias="json")  # json content

    def __init__(self, Json: Union[dict, List], **kwargs):
        """
        Construct Json component

        :param json: json content
        """
        super().__init__(Json=Json, **kwargs)


class App(BaseMessageComponent):
    """
    App component
    Available for outbound message
    """
    type = MessageComponentTypes.App
    content: str  # app content

    def __init__(self, content: str, **kwargs):
        """
        Construct App component

        :param content: app content
        """
        super().__init__(content=content, **kwargs)


class Unknown(BaseMessageComponent):
    """
    Unknown component
    Not a valid component for outbound message
    """
    type = MessageComponentTypes.Unknown
    text: str  # content in string

    def __str__(self):
        return ''

    def __repr__(self):
        return '[Unknown]'


class ComponentTypes(Enum):
    Plain = Plain
    Source = Source
    At = At
    AtAll = AtAll
    Face = Face
    Image = Image
    Quote = Quote
    Xml = Xml
    Json = Json
    App = App
    Unknown = Unknown


MessageComponents = {
    'At':      At,
    'AtAll':   AtAll,
    'Face':    Face,
    'Plain':   Plain,
    'Image':   Image,
    'Source':  Source,
    'Quote':   Quote,
    'Xml':     Xml,
    'Json':    Json,
    'App':     App,
    'Unknown': Unknown
}


class BotMessage(BaseModel):
    type: str = 'BotMessage'
    messageId: int


class ImageType(Enum):
    Friend = 'friend'
    Group = 'group'



class MiraiException(Exception):
    """
    Generic exception
    """
    pass


class NetworkException(MiraiException):
    """
    Exceptions such as not able to reach mirai, possibly due to service not started or incorrect ip/port
    """
    pass


class AuthenticationException(MiraiException):
    """
    Exceptions such as incorrect auth_key
    """
    pass


class SessionException(MiraiException):
    """
    Exceptions such as incorrect session_key
    """
    pass


class PrivilegeException(MiraiException):
    """
    You do not have permission to perform such action (such as mute, kick, etc.)
    """
    pass


class UnknownTargetException(MiraiException):
    """
    Target not found (sending to non-existing group, friend, etc.)
    """
    pass


class BadRequestException(MiraiException):
    """
    Request contains invalid parameter
    """
    pass


class ServerException(MiraiException):
    """
    Server is returning non 200 return code, see console log for more details
    This is more likely to be mirai side issue, but sometimes incorrect parameters can cause this problem too
    """
    pass



from typing import Union, List, Type, Dict
from datetime import timedelta
from pathlib import Path
import json
from .log import create_logger

from .models.message import BotMessage, ImageType, MessageChain, \
    Source, Image, Quote, Plain, BaseMessageComponent, LocalImage
from .models.events import *
from .models.entity import Friend, Group, GroupSetting, Member, MemberChangeableSetting
from .network import HttpClient
from .exceptions import AuthenticationException, MiraiException, NetworkException, SessionException

__ALL__ = [
    'Bot'
]


__logger = create_logger('Bot')


def retry_once(func):
    async def wrapper(self, *args, **kwargs):
        try:
            return await func(self, *args, **kwargs)
        except (NetworkException, SessionException, AuthenticationException):
            __logger.exception('Trying handshake due to the following exception')
        try:
            await self.handshake()
            return await func(self, *args, **kwargs)
        except (NetworkException, SessionException, AuthenticationException):
            __logger.exception('Unable to handshake')
        return None

    return wrapper


class Bot:
    """
    See https://github.com/mamoe/mirai-api-http for details
    """

    def __init__(self, qq: int, host: str = '127.0.0.1', port: int = 8080, auth_key: str = 'abcdefgh', loop=None):
        self.qq = qq
        self.auth_key = auth_key
        self.base_url = f'http://{host}:{port}'
        self.loop = loop
        self.session = HttpClient(self.base_url, loop=self.loop)
        self.session_key = ''

    async def handshake(self):
        """
        Authenticate and verify the session_key
        Automatically called if session_key needs to be updated
        """
        await self.auth()
        await self.verify()

    async def auth(self) -> None:
        """
        Post auth_key, and get session_key
        """
        result = await self.session.post('/auth', data={'authKey': self.auth_key})
        self.session_key = result.get('session')

    async def verify(self) -> None:
        """
        Post session_key to verify the session
        """
        await self.session.post('/verify',
                                data={
                                    'sessionKey': self.session_key,
                                    'qq':         self.qq
                                })

    async def release(self) -> None:
        """
        Post session_key to release the session
        Needs to be called manually if Updater is not used
        """
        await self.session.post('/release',
                                data={
                                    'sessionKey': self.session_key,
                                    'qq':         self.qq
                                })

    @staticmethod
    def _handle_target_as(target: Union[Group, Friend, Member, int],
                          as_type: Union[Type[Group], Type[Friend], Type[Member]]):
        """
        Internal use only, convert target to id

        :param target: Union[Group, Friend, Member, int]
        :param as_type: Group, Friend or Member
        :return: id, int
        """
        if isinstance(target, int):
            return target
        elif isinstance(target, as_type):
            return target.id
        else:
            raise ValueError(f'Invalid target as {type(as_type)} object.')

    @retry_once
    async def send_friend_message(self,
                                  friend: Union[Friend, int],
                                  message: Union[
                                      MessageChain,
                                      BaseMessageComponent,
                                      List[BaseMessageComponent],
                                      str
                                  ],
                                  quote_source: Union[int, Source] = None) -> BotMessage:
        """
        Send friend message

        :param friend: int or Friend object as target
        :param message: MessageChain, BaseMessageComponent, List of BaseMessageComponent or str, the content to send
        :param quote_source: int (the 64-bit int) or Source, the message to quote
        :return: BotMessage (contains message id)
        """
        data = {
            'sessionKey':   self.session_key,
            'target':       Bot._handle_target_as(friend, Friend),
            'messageChain': await self._handle_message_chain(message, Friend)
        }
        if quote_source:
            if isinstance(quote_source, int):
                data['quote'] = quote_source
            elif isinstance(quote_source, Source):
                data['quote'] = quote_source.id
        result = await self.session.post('/sendFriendMessage', data=data)

        return BotMessage.parse_obj(result)

    @retry_once
    async def send_group_message(self,
                                 group: Union[Group, int],
                                 message: Union[
                                     MessageChain,
                                     BaseMessageComponent,
                                     List[BaseMessageComponent],
                                     str
                                 ],
                                 quote_source: Union[int, Source] = None) -> BotMessage:
        """
        Send group message

        :param group: int or Group object as target
        :param message: MessageChain, BaseMessageComponent, List of BaseMessageComponent or str, the content to send
        :param quote_source: int (the 64-bit int) or Source, the message to quote
        :return: BotMessage (contains message id)
        """
        data = {
            'sessionKey':   self.session_key,
            'target':       Bot._handle_target_as(group, Group),
            'messageChain': await self._handle_message_chain(message, Group)
        }
        if quote_source:
            if isinstance(quote_source, int):
                data['quote'] = quote_source
            elif isinstance(quote_source, Source):
                data['quote'] = quote_source.id
        result = await self.session.post('/sendGroupMessage', data=data)
        return BotMessage.parse_obj(result)

    @retry_once
    async def recall(self, source: Union[Source, int]) -> None:
        """
        Recall a message
        Success if no exception is raised

        :param source: int (the 64-bit int) or Source
        """
        data = {
            'sessionKey': self.session_key,
        }
        if isinstance(source, int):
            data['target'] = source
        elif isinstance(source, Source):
            data['target'] = source.id
        else:
            raise MiraiException('Invalid source argument')

        await self.session.post('/recall', data=data)

    @property
    @retry_once
    async def groups(self) -> List[Group]:
        """
        Get list of joined groups

        :return: List of Group
        """
        params = {
            'sessionKey': self.session_key,
        }
        result = await self.session.get('/groupList', params=params)
        return [Group.parse_obj(group_info) for group_info in result]

    @property
    @retry_once
    async def friends(self) -> List[Friend]:
        """
        Get list of friends

        :return: List of Friend
        """
        params = {
            'sessionKey': self.session_key,
        }
        result = await self.session.get('/friendList', params=params)
        return [Friend.parse_obj(friend_info) for friend_info in result]

    @retry_once
    async def get_members(self, target: Union[Group, int]) -> List[Member]:
        """
        Get list of members of a group

        :param target: int or Group, the target group
        :return: List of Member
        """
        if isinstance(target, int):
            group = target
        else:
            group = target.id
        params = {
            'sessionKey': self.session_key,
            'target':     group
        }
        result = await self.session.get('/memberList', params=params)
        return [Member.parse_obj(member_info) for member_info in result]

    @retry_once
    async def upload_image(self, image_type: ImageType, image_path: Union[Path, str]) -> Optional[Image]:
        """
        Upload a image to QQ server. The image between group and friend is not exchangeable
        This function can be called separately to acquire image uuids, or automatically if using LocalImage while sending

        :param image_type: ImageType, Friend or Group
        :param image_path: absolute path of the image
        :return: Image object
        """
        if isinstance(image_path, str):
            image_path = Path(image_path)

        if not image_path.exists():
            raise FileNotFoundError('Image not found.')

        data = {
            'sessionKey': self.session_key,
            'type':       image_type.value
        }
        result = await self.session.upload('/uploadImage', file=image_path, data=data)
        return Image.parse_obj(result)

    @retry_once
    async def fetch_message(self, count: int) -> List[Event]:
        """
        Fetch a list of messages
        This function is called automatically if using polling instead of websocket

        :param count: maximum count of one fetch
        :return: List of Event
        """
        params = {
            'sessionKey': self.session_key,
            'count':      count
        }
        result = await self.session.get('/fetchMessage', params=params)

        try:
            for index in range(len(result)):
                result[index] = self._parse_event(result[index])
        except:
            global __logger
            __logger.exception('Unhandled exception')
        return result

    @retry_once
    async def mute_all(self, group: Union[Group, int]) -> None:
        """
        Mute all non admin members in group
        Success if no exception is raised

        :param group: int or Group, the target group
        """
        params = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group)
        }

        await self.session.get('/muteAll', params=params)

    @retry_once
    async def unmute_all(self, group: Union[Group, int]) -> None:
        """
        Unmute all non admin members in group
        Success if no exception is raised

        :param group: int or Group, the target group
        """
        params = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group)
        }

        await self.session.get('/unmuteAll', params=params)

    @retry_once
    async def get_member_info(self, group: Union[Group, int], member: Union[Member, int]) -> MemberChangeableSetting:
        """
        Get the info of a member

        :param group: int or Group, target group
        :param member: int or Member, target member
        :return: MemberChangeableSetting
        """
        params = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
            'memberId':   self._handle_target_as(target=member, as_type=Member)
        }

        result = await self.session.get('/memberInfo', params=params)
        return MemberChangeableSetting.parse_obj(result)

    @retry_once
    async def get_bot_member_info(self, group: Union[Group, int]) -> MemberChangeableSetting:
        """
        Get the info of this bot

        :param group: int or Group, target group
        :return: MemberChangeableSetting
        """
        return await self.get_member_info(group, self.qq)

    @retry_once
    async def set_member_info(self, group: Union[Group, int],
                              member: Union[Member, int],
                              setting: MemberChangeableSetting) -> None:
        """
        Set the info of a member
        Success if no exception is raised

        :param group: int or Group, target group
        :param member: int or Member, target member
        :param setting: MemberChangeableSetting, the new settings
        """
        data = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
            'memberId':   self._handle_target_as(target=member, as_type=Member),
            'info':       json.loads(setting.json())
        }

        await self.session.post('/memberInfo', data=data)

    @retry_once
    async def get_group_config(self, group: Union[Group, int]) -> GroupSetting:
        """
        Get the group config of a group

        :param group: int or Group, target group
        :return: GroupSetting
        """
        params = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
        }
        result = await self.session.get('/groupConfig', params=params)
        return GroupSetting.parse_obj(result)

    @retry_once
    async def set_group_config(self, group: Union[Group, int],
                               config: GroupSetting) -> None:
        """
        Set the group config of a group
        Success if no exception is raised

        :param group: int or Group, target group
        :param config: GroupSetting
        """
        data = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
            'config':     json.loads(config.json())
        }

        await self.session.post('/groupConfig', data=data)

    @retry_once
    async def mute(self, group: Union[Group, int],
                   member: Union[Member, int],
                   time: Union[timedelta, int]) -> None:
        """
        Mute a member of a group
        Success if no exception is raised

        :param group: int or Group, target group
        :param member: int or Member, target member
        :param time: int or datetime.timedelta, must between 1 minutes and 30 days
        """
        if isinstance(time, timedelta):
            time = int(time.total_seconds())
        time = min(86400 * 30, max(60, time))  # time should between 1 minutes and 30 days
        data = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
            'MemberId':   Bot._handle_target_as(target=member, as_type=Member),
            'time':       time
        }
        await self.session.post('/mute', data=data)

    @retry_once
    async def unmute(self, group: Union[Group, int],
                     member: Union[Member, int]) -> None:
        """
        Unmute a member of a group
        Success if no exception is raised

        :param group: int or Group, target group
        :param member: int or Member, target member
        """
        data = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
            'MemberId':   Bot._handle_target_as(target=member, as_type=Member)
        }
        await self.session.post('/unmute', data=data)

    @retry_once
    async def kick(self, group: Union[Group, int],
                   member: Union[Member, int],
                   message: str = '') -> None:
        """
        Kick a member of a group

        :param group: int or Group, target group
        :param member: int or Member, target member
        :param message: string, message to the member
        """
        data = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
            'MemberId':   Bot._handle_target_as(target=member, as_type=Member)
        }
        if message:
            data['msg'] = message

        await self.session.post('/kick', data=data)

    @retry_once
    async def respond_request(self,
                              request: Union[NewFriendRequestEvent, MemberJoinRequestEvent],
                              response: Union[NewFriendRequestResponse, MemberJoinRequestResponse],
                              message: str = ''):
        if isinstance(request, NewFriendRequestEvent):
            assert isinstance(response, (NewFriendRequestResponse, int)), f'Response type mismatch'
            response = response.value if isinstance(response, NewFriendRequestResponse) else response
            data = {
                'sessionKey': self.session_key,
                'eventId': request.requestId,
                'fromId':  request.supplicant,
                'groupId': request.sourceGroup,
                'operate': response,
                'message': message
            }
            return await self.session.post('/resp/newFriendRequestEvent', data=data)
        elif isinstance(request, MemberJoinRequestEvent):
            assert isinstance(response, (MemberJoinRequestResponse, int)), f'Response type mismatch'
            response = response.value if isinstance(response, MemberJoinRequestResponse) else response
            data = {
                'sessionKey': self.session_key,
                'eventId': request.requestId,
                'fromId':  request.supplicant,
                'groupId': request.sourceGroup,
                'operate': response,
                'message': message
            }
            return await self.session.post('/resp/memberJoinRequestEvent', data=data)
        else:
            raise TypeError(f'Unsupported event: {str(request)}')

    async def _handle_image(self, message: BaseMessageComponent, image_type: ImageType) -> dict:
        """
        Internal use only
        Convert LocalImage to Image, and everything to json

        :param message: BaseMessageComponent
        :param image_type: ImageType
        :return: json representation
        """
        if not isinstance(message, LocalImage):
            return json.loads(message.json())

        image = await self.upload_image(image_type, message.path)

        return {
            'type':    'Image',
            'imageId': image.imageId
        }

    async def _handle_message_chain(self, message: Union[
        MessageChain,
        BaseMessageComponent,
        List[BaseMessageComponent],
        str
    ], as_type: Union[Type[Group], Type[Friend]]) -> List:
        """
        Internal use only
        Convert MessageChain to json

        :param message: MessageChain
        :param as_type: Group or Friend
        :return: list
        """
        if isinstance(message, MessageChain):
            return json.loads(message.json())
        elif isinstance(message, str):
            return [json.loads(Plain(text=message).json())]
        elif isinstance(message, BaseMessageComponent):
            if as_type == Group:
                image_type = ImageType.Group
            else:
                image_type = ImageType.Friend
            return [await self._handle_image(message, image_type)]
        elif isinstance(message, (tuple, list)):
            if as_type == Group:
                image_type = ImageType.Group
            else:
                image_type = ImageType.Friend
            result = [await self._handle_image(m, image_type=image_type) for m in message]
            return result
        else:
            raise ValueError('Invalid message')

    @retry_once
    async def get_config(self) -> dict:
        """
        Get the config of http api

        :return: config
        """
        params = {
            'sessionKey': self.session_key
        }

        result = await self.session.get('/config', params=params)
        return result

    @retry_once
    async def set_config(self, cache_size: int = 4096, enable_websocket: bool = True) -> None:
        """
        Set the config of http api
        Success if no exception is raised

        :param cache_size: int, the size of message cache
        :param enable_websocket: bool, whether to enable websocket (will disable fetch_message accordingly)
        """
        data = {
            'sessionKey':      self.session_key,
            'cacheSize':       cache_size,
            'enableWebsocket': enable_websocket
        }

        await self.session.post('/config', data=data)

    def _parse_event(self, result) -> Event:
        """
        Internal use only
        Parse event or message from json to Event

        :param result: the json
        :return: Event
        """
        if hasattr(Events, result['type']):  # if Event
            try:
                if 'messageChain' in result:  # construct message chain
                    # parse quote first
                    first_component = result['messageChain'][1]
                    if first_component['type'] == 'Quote':  # FIXME: add the first two message part back
                        result['messageChain'][1]['origin'] = MessageChain.custom_parse(
                            result['messageChain'][1]['origin'])
                        try:
                            del result['messageChain'][2]  # delete duplicated at
                            if result['messageChain'][2]['type'] == 'Plain' and result['messageChain'][2]['text'] == ' ':
                                del result['messageChain'][2]  # delete space after duplicated at
                        except:
                            global __logger
                            __logger.exception('Please open a github issue to report this error')
                    # for idx, component in enumerate(result['messageChain']):
                    #     if component['type'] == 'Quote':
                    #         result['messageChain'][idx]['origin'] = MessageChain.custom_parse(
                    #             result['messageChain'][idx]['origin'])
                    result['messageChain'] = MessageChain.custom_parse(result['messageChain'])
                result = Events[result['type']].value.parse_obj(result)
            except:
                global __logger
                __logger.exception('Unhandled exception')
            return result
        else:
            raise ValueError('Invalid message chain')

    def _websocket_handler(self, handler: callable) -> callable:
        """
        Internal use only
        Wrap the handler, and convert json to Event

        :param handler: callable, the handler
        :return: wrapped handler
        """

        async def _handler(result: Union[List, Dict]):
            """
            an example handler for create_websocket
            :param result: json
            """
            result = self._parse_event(result)
            await handler(result)

        return _handler

    @retry_once
    async def create_websocket(self, handler, ws_close_handler=None, listen: str = 'all') -> None:
        """
        Register callback for websocket. Once an Event or Message is received, the handler will be invoked

        :param handler: callable
        :param ws_close_handler: callable, websocket shutdown hook
        :param listen: 'all', 'event' or 'message'
        """
        if listen not in ('all', 'event', 'message'):
            raise ValueError("listen must be one of 'all', 'event' or 'message'")
        if ws_close_handler is None:
            async def ws_close_handler(event):
                pass
        await self.session.websocket(f'/{listen}?sessionKey={self.session_key}',
                                     self._websocket_handler(handler), ws_close_handler)



import asyncio
from typing import DefaultDict, Union, List, Callable, Any, Awaitable
from collections import defaultdict
from dataclasses import dataclass
import signal
from .log import create_logger, install_logger

from .bot import Bot
from .models.events import Event, EventTypes
from .exceptions import SessionException, NetworkException, AuthenticationException, ServerException


class Updater:
    def __init__(self, bot: Bot, use_websocket: bool = True):
        """
        Initialize Updater

        :param bot: the Bot object to use
        :param use_websocket: bool. whether websocket (recommended) should be used
        """
        self.bot = bot
        self.loop = bot.loop
        self.logger = create_logger('Updater')
        self.event_handlers: DefaultDict[EventTypes, List[EventHandler]] = defaultdict(lambda: list())
        self.use_websocket = use_websocket

    async def run_task(self, shutdown_hook: callable = None):
        """
        return awaitable coroutine to run in event loop (must be the same loop as bot object)

        :param shutdown_hook: callable, if running in main thread, this must be set. Trigger is called on shutdown
        """
        self.logger.debug('Run tasks')
        tasks = [
            self.handshake()
        ]
        if not self.use_websocket:
            tasks.append(self.message_polling())
        if shutdown_hook:
            tasks.append(self.raise_shutdown(shutdown_hook))
        await asyncio.wait(tasks)

    def add_handler(self, event: Union[EventTypes, List[EventTypes]]):
        """
        Decorator for event listeners
        Catch all is not supported at this time

        :param event: events.Events
        """
        def receiver_wrapper(func):
            if not asyncio.iscoroutinefunction(func):
                raise TypeError("event body must be a coroutine function.")

            # save function and its parameter types
            event_handler = EventHandler(func)
            if isinstance(event, EventTypes):
                # add listener
                self.event_handlers[event].append(event_handler)
            else:
                for e in event:
                    if isinstance(e, EventTypes):
                        self.event_handlers[e].append(event_handler)
            return func

        return receiver_wrapper

    def run(self, log_to_stderr=True) -> None:
        """
        Start the Updater and block the thread

        :param log_to_stderr: if you are setting other loggers that capture the log from this Library, set to False
        """
        asyncio.set_event_loop(self.loop)

        shutdown_event = asyncio.Event()

        def _signal_handler(*_: Any) -> None:
            shutdown_event.set()

        try:
            self.loop.add_signal_handler(signal.SIGTERM, _signal_handler)
            self.loop.add_signal_handler(signal.SIGINT, _signal_handler)
        except (AttributeError, NotImplementedError):
            pass

        if log_to_stderr:
            install_logger()

        self.loop.create_task(self.run_task(shutdown_hook=shutdown_event.wait))
        self.loop.run_forever()

    async def handshake(self):
        """
        Internal use only, automatic handshake
        Called when launch or websocket disconnects

        :return:
        """
        while True:
            try:
                await self.bot.handshake()
                if self.use_websocket:
                    asyncio.run_coroutine_threadsafe(
                        self.bot.create_websocket(self.event_caller, self.handshake), self.loop)
                return
            except NetworkException:
                self.logger.warning('Unable to communicate with Mirai console, retrying in 5 seconds')
                await asyncio.sleep(5)
            except Exception as e:
                self.logger.exception(f'retrying in 5 seconds')
                await asyncio.sleep(5)

    async def message_polling(self, count=5, interval=0.5) -> None:
        """
        Internal use only, polling message and fire events

        :param count: maximum message count for each polling
        :param interval: minimum interval between two polling
        """
        while True:
            await asyncio.sleep(interval)
            try:
                results: List[Event] = await self.bot.fetch_message(count)
                if len(results) > 0:
                    self.logger.debug('Received messages:\n' + '\n'.join([str(result) for result in results]))
                for result in results:
                    asyncio.run_coroutine_threadsafe(self.event_caller(result), self.loop)
            except Exception as e:
                self.logger.warning(f'{e}, new handshake initiated')
                await self.handshake()

    async def event_caller(self, event: Event) -> None:
        """
        Internal use only, call the event handlers sequentially

        :param event: the event
        """
        for handler in self.event_handlers[event.type]:
            if await handler.func(event):  # if the function returns True, stop calling next event
                break

    async def raise_shutdown(self, shutdown_event: Callable[..., Awaitable[None]]) -> None:
        """
        Internal use only, shutdown

        :param shutdown_event: callable
        """
        await shutdown_event()
        await self.bot.release()
        raise Shutdown()


@dataclass
class EventHandler:
    """
    Contains the callback function
    """
    func: Callable


class Shutdown(Exception):
    """
    Internal use only
    Shutdown Event
    """
    pass
